package com.amazonaws.samples;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import com.amazonaws.ClientConfiguration;
import com.amazonaws.regions.*;
import com.amazonaws.regions.Region;
import com.amazonaws.services.s3.*;
import com.amazonaws.services.s3.model.*;

public class LowLevel_LargeObjectCopy {

    private static String env(String name, String def) {
        String v = System.getenv(name);
        return (v != null) ? v : def;
    }

    private static int envInt(String name, int def) {
        String v = System.getenv(name);
        return (v != null) ? Integer.valueOf(v) : def;
    }

    private static boolean envBoolean(String name, boolean def) {
        String v = System.getenv(name);
        return (v != null) ? Boolean.valueOf(v) : def;
    }

    public static void main(String[] args) throws IOException {
        System.out.println(new SimpleDateFormat("yyyy/MM/dd HH:mm:ss").format(new Date()));

        // Example settings via env variables:
        // S3_ENDPOINT="http://backup42.docker.tx.caringo.com:4285" S3_ACCESS_KEY=a33898e802a42d5e099536b8a9727a3d S3_SECRET_KEY=secret mvn clean compile exec:java

        String endpoint = env("S3_ENDPOINT", "https://support.cloud.caringo.com"); // your domain
        String bucketName = env("S3_BUCKET", "tmp"); // your domain
        // These uploader user credentials will expire periodically, please contact Caringo Support for latest.
        System.setProperty("aws.accessKeyId", env("S3_ACCESS_KEY", "6bbfc09e5f901061805c0e3fa77f18ff"));
        System.setProperty("aws.secretKey", env("S3_SECRET_KEY", "g9Dum60C1IjSGytKjh2kfSBnEC9YYInXo98fFCld"));

        int PART_SIZE_MB = envInt("S3_PART_SIZE", 2000 /* 2GB */);
        int CONNECTION_TIMEOUT_MS = envInt("S3_CONNECTION_TIMEOUT_MS", 240000);
        int SOCKET_TIMEOUT_MS = envInt("S3_SOCKET_TIMEOUT_MS", 240000);
        boolean enforceV4 = envBoolean("S3_ENFORCE_V4", true);
        boolean bucketInPath = envBoolean("S3_BUCKET_IN_PATH", true); // default is bucket in host, https://mybucket.support.cloud.caringo.com/object
        if (enforceV4) {
            System.setProperty("com.amazonaws.services.s3.enforceV4", "true"); // enforce V4 signature auth
        }
        ClientConfiguration cfg = new ClientConfiguration();
        System.out.println("Default getConnectionTimeout=" + cfg.getConnectionTimeout() + "ms, getSocketTimeout=" + cfg.getSocketTimeout() + "ms");
        cfg.setConnectionTimeout(CONNECTION_TIMEOUT_MS );
        cfg.setSocketTimeout(SOCKET_TIMEOUT_MS);
        S3ClientOptions s3opts = new S3ClientOptions();
        s3opts.setPathStyleAccess(bucketInPath);
        AmazonS3Client s3Client = new AmazonS3Client(cfg);
        s3Client.setS3ClientOptions(s3opts);
        if (enforceV4) {
            Region region = com.amazonaws.regions.Region.getRegion(Regions.US_EAST_1);
            s3Client.setRegion(region);
            s3Client.setSignerRegionOverride(region.getName()); // fixes "Invalid S3 URI" from AWS "s3.*" check
        }
        System.out.println("Copying " + endpoint + ", part size=" + PART_SIZE_MB + "MB, getConnectionTimeout=" + cfg.getConnectionTimeout() + "ms, getSocketTimeout=" + cfg.getSocketTimeout() + "ms, enforceV4="+enforceV4 + ", bucketInPath="+bucketInPath);
        s3Client.setEndpoint(endpoint);

        String sourceBucketName = bucketName;
        String targetBucketName = bucketName;
        String sourceObjectKey  = "samples/5G";
        String targetObjectKey  = "output/uploader-swarm/LowLevel_LargeObjectCopy-5G";
        System.out.println("Copying " + endpoint + "/" + sourceBucketName + "/" + sourceObjectKey + " to /" + targetBucketName + "/" + targetObjectKey + ", part size=" + PART_SIZE_MB + "MB, getConnectionTimeout=" + cfg.getConnectionTimeout() + "ms, getSocketTimeout=" + cfg.getSocketTimeout() + "ms, enforceV4="+enforceV4 + ", bucketInPath="+bucketInPath);

        // List to store copy part responses.

        List<CopyPartResult> copyResponses =
                new ArrayList<CopyPartResult>();

        InitiateMultipartUploadRequest initiateRequest =
                new InitiateMultipartUploadRequest(targetBucketName, targetObjectKey);

        InitiateMultipartUploadResult initResult =
                s3Client.initiateMultipartUpload(initiateRequest);

        try {
            // Get object size.
            GetObjectMetadataRequest metadataRequest =
                    new GetObjectMetadataRequest(sourceBucketName, sourceObjectKey);

            ObjectMetadata metadataResult = s3Client.getObjectMetadata(metadataRequest);
            long objectSize = metadataResult.getContentLength(); // in bytes

            // Copy parts.
            long partSize = PART_SIZE_MB * (long)Math.pow(2.0, 20.0); // MB

            long bytePosition = 0;
            for (int i = 1; bytePosition < objectSize; i++)
            {
                long lastByte = bytePosition + partSize -1 >= objectSize ? objectSize - 1 : bytePosition + partSize - 1;
                long start = System.nanoTime();
                System.out.println("Copying part " + i + " from " + bytePosition + "-" + lastByte + " ...");
                CopyPartRequest copyRequest = new CopyPartRequest()
                        .withDestinationBucketName(targetBucketName)
                        .withDestinationKey(targetObjectKey)
                        .withSourceBucketName(sourceBucketName)
                        .withSourceKey(sourceObjectKey)
                        .withUploadId(initResult.getUploadId())
                        .withFirstByte(bytePosition)
                        .withLastByte(lastByte)
                        .withPartNumber(i);
                copyResponses.add(s3Client.copyPart(copyRequest));
                System.out.println("Finished copying part " + i + " from " + bytePosition + "-" + lastByte + " in " + (System.nanoTime() - start) / 1000 / 1000 + "ms");
                bytePosition += partSize;

            }
            CompleteMultipartUploadRequest completeRequest = new
                    CompleteMultipartUploadRequest(
                    targetBucketName,
                    targetObjectKey,
                    initResult.getUploadId(),
                    GetETags(copyResponses));

            CompleteMultipartUploadResult completeUploadResponse =
                    s3Client.completeMultipartUpload(completeRequest);
        } catch (Exception e) {
            System.out.println(new SimpleDateFormat("yyyy/MM/dd HH:mm:ss").format(new Date()));
            System.out.println(e.getMessage());
            System.exit(1);
        }
    }

    // Helper function that constructs ETags.
    static List<PartETag> GetETags(List<CopyPartResult> responses)
    {
        List<PartETag> etags = new ArrayList<PartETag>();
        for (CopyPartResult response : responses)
        {
            etags.add(new PartETag(response.getPartNumber(), response.getETag()));
        }
        return etags;
    }
}
